﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Collections.Generic;
using System.IO;

namespace AppReviews.Web
{
    public class WebApi
    {
        internal const string AppDataTemplate = "http://catalog.zune.net/v3.2/{0}/apps/{1}?version=latest&clientType=WinMobile%207.1&store=Zest";
        internal const string ReviewTemplate = "http://catalog.zune.net/v3.2/{0}/apps/{1}/reviews";
        internal const string SearchTemplate = "http://catalog.zune.net/v3.2/en-us/apps?q={0}&clientType=WinMobile%207.1&store=zest";
        internal const string TranslateTemplate = "http://api.microsofttranslator.com/V2/Http.svc/Translate?appId=70B153FB80F37BC6D47A90ACF245357FAB682C44&to=en&text={0}";

        private static List<WebWorkItem> _queue = new List<WebWorkItem>();
        private static int _isProcessingRequest = 0;

        private static WebWorkItem QueueGetString(string uri, object userData, object groupId, WebWorkItem.CompletedHandler completed)
        {
            WebWorkItem result = new WebWorkItem()
            {
                Uri = uri,
                UserData = userData,
                Completed = completed,
                GroupId = groupId
            };

            Add(result);
            return result;
        }

        public static WebWorkItem QueueTranslate(string translateFrom, object userData, object groupId, WebWorkItem.CompletedHandler completed)
        {
            return QueueGetString(string.Format(TranslateTemplate, HttpUtility.UrlEncode(translateFrom)), userData, groupId, completed);
        }

        public static WebWorkItem QueueAppSearch(string searchTerm, object userData, object groupId, WebWorkItem.CompletedHandler completed)
        {
            return QueueGetString(string.Format(SearchTemplate, HttpUtility.UrlEncode(searchTerm)), userData, groupId, completed);
        }

        public static WebWorkItem QueueGetAppData(string marketLocale, string appId, object userData, object groupId, WebWorkItem.CompletedHandler completed)
        {
            return QueueGetString(string.Format(AppDataTemplate, marketLocale, appId), userData, groupId, completed);
        }

        public static WebWorkItem QueueGetAppReviews(string marketLocale, string appId, object userData, object groupId, WebWorkItem.CompletedHandler completed)
        {
            return QueueGetString(string.Format(ReviewTemplate, marketLocale, appId), userData, groupId, completed);
        }


        private static object _groupIdToMoveToTop = null;

        /// <summary>
        /// Move all requests for a specified group to the top of the queue
        /// </summary>
        /// <param name="groupId"></param>
        public static void MoveToTop(object groupId)
        {
            _groupIdToMoveToTop = groupId;
            if (groupId == null) return;

            lock (_queue)
            {
                int count = _queue.Count;
                int indexToInsertAt = 0;
                for (int i = 0; i < count; i++)
                {
                    WebWorkItem item = _queue[i];
                    if (item.GroupId == null)
                    {
                        indexToInsertAt++;
                    }
                    else
                    {
                        break;
                    }
                }

                    for (int i = 0; i < count; i++)
                {
                    WebWorkItem item = _queue[i];
                    if (item.GroupId != groupId)
                    {
                        continue;
                    }

                    _queue.RemoveAt(i);
                    _queue.Insert(indexToInsertAt, item);
                    indexToInsertAt++;
                }
            }
        }

        /// <summary>
        /// Cancels a given request
        /// </summary>
        /// <param name="item">request to cancel</param>
        public static void Cancel(WebWorkItem item)
        {
            bool itemFound = false;
            int queueCountAtEnd = 0;

            lock (_queue)
            {
                for (int i = 0; i < _queue.Count; i++)
                {
                    WebWorkItem queuedItem = _queue[i];
                    if (queuedItem == item)
                    {
                        _queue.RemoveAt(i);
                        queueCountAtEnd = _queue.Count;
                        break;
                    }
                }
            }

            if (itemFound)
            {
                if (queueCountAtEnd == 0)
                {
                    Globals.IsLoading = false;
                }

                if (item.Completed != null)
                {
                    item.Result = null;
                    item.Completed(item);
                }
            }
        }

        public static void CancelAll()
        {
            WebWorkItem[] items;
            lock (_queue)
            {
                if (_queue.Count == 0)
                {
                    return;
                }

                items = new WebWorkItem[_queue.Count];
                _queue.CopyTo(items);
                _queue.Clear();
            }

            foreach (WebWorkItem item in items)
            {
                if (item.Completed != null)
                {
                    item.Result = null;
                    item.Completed(item);
                }
            }

            Globals.IsLoading = false;
        }

        private static void ProcessRequestIfPossible()
        {
            if (_isProcessingRequest >= 6) return;
            // see if there's anything else in the queue
            int count;
            WebWorkItem item = null;

            lock (_queue)
            {
                count = _queue.Count;
                if (count > 0)
                {
                    item = _queue[0];
                    _queue.RemoveAt(0);
                }
            }

            if (count == 0)
            {
                if (_isProcessingRequest == 0)
                {
                    Globals.IsLoading = false;
                    Globals.StatusText = String.Empty;
                }
                return;
            }

            Globals.StatusText = string.Format("{0} items left", count + _isProcessingRequest);

            _isProcessingRequest++;
            HttpWebRequest request = (HttpWebRequest)WebRequest.Create(item.Uri);
            request.BeginGetResponse(BeginResponse, new HttpWebRequestData() { Request = request, Item = item });
        }

        /// <summary>
        /// Parses response from server and signals the async operation
        /// </summary>
        /// <param name="ar"></param>
        private static void BeginResponse(IAsyncResult ar)
        {
            HttpWebRequestData data = (HttpWebRequestData)ar.AsyncState;
            string response = null;
            Exception exception = null;

            try
            {
                using (WebResponse webResponse = data.Request.EndGetResponse(ar))
                {
                    using (Stream stream = webResponse.GetResponseStream())
                    {
                        byte[] buffer = new byte[stream.Length];
                        stream.Read(buffer, 0, (int)stream.Length);
                        if ((buffer.Length >= 3) && (buffer[0] == 0xef) && (buffer[1] == 0xbb)
                            && (buffer[2] == 0xbf))
                        {
                            response = System.Text.Encoding.UTF8.GetString(buffer, 3, buffer.Length - 3);
                        }
                        else
                        {
                            response = System.Text.Encoding.UTF8.GetString(buffer, 0, buffer.Length);
                        }
                        stream.Close();
                    }
                    webResponse.Close();
                }
            }
            catch (Exception e)
            {
                exception = e;
            }

            WebWorkItem item = data.Item;
            if (item.Completed != null)
            {
                item.Result = response;
                item.Error = exception;
                item.Completed(item);
            }

            _isProcessingRequest--;
            ProcessRequestIfPossible();
        }

        private static void Add(WebWorkItem item)
        {
            lock (_queue)
            {
                if ((item.GroupId != null) && (item.GroupId == _groupIdToMoveToTop))
                {
                    int count = _queue.Count;
                    
                    bool isAdded = false;
                    for (int i = 0; i < count; i++)
                    {
                        WebWorkItem existingItem = _queue[i];
                        if ((existingItem.GroupId == item.GroupId) || (existingItem.GroupId == null))
                        {
                            continue;
                        }

                        _queue.Insert(i, item);
                        isAdded = true;
                        break;
                    }

                    if (!isAdded)
                    {
                        _queue.Add(item);
                    }
                }
                else
                {
                    if (item.GroupId == null) // if exception
                    {
                        _queue.Insert(0, item);
                    }
                    else
                    {
                        _queue.Add(item);
                    }
                }
            }

            Globals.IsLoading = true;
            ProcessRequestIfPossible();
        }
    }
}
